from __future__ import absolute_import, print_function
import gzip
import json
from collections import OrderedDict
import requests

try:  # pragma: no cover
    # For Python 3.0 and later
    from urllib.request import urlopen, Request, URLError
    from urllib.error import HTTPError
    from io import StringIO
    from urllib.parse import urlencode
except ImportError:  # pragma: no cover
    # Fall back to Python 2
    from urllib import urlencode
    from urllib2 import urlopen, Request, URLError, HTTPError
    from StringIO import StringIO
    from types import ListType as list

from centric8python.exceptions import SessionError, NoResultError
from centric8python.xml_node_operations import get_operations_root, ChangeNode, ChangeAttribute, tostring, CreateNode


def xml_to_string(xml):
    try:
        return tostring(xml).decode('utf-8')

    except TypeError:
        return tostring(xml)


def load_json(data):
    try:
        return json.loads(data)
    except TypeError:
        return json.loads(data.decode('utf-8'))


def authenticate(func):
    def authenticate_and_call(*args, **kwargs):
        client = args[0]
        if not client.is_logged_in:
            raise SessionError('Not logged in')
        return func(*args, **kwargs)

    return authenticate_and_call


def exists(path):  # pragma: no cover
    try:
        response = urlopen(path)
        return response.status == 200
    except IOError:
        return False
    except AttributeError:
        return response.code == 200
    except Exception as e:
        print(e)


def check_status(dom):  # pragma: no cover
    status = dom.xpath('Status')[0]
    if status.text == 'Failed':
        error = dom.xpath('Error')[0]
        if error.text == 'SRV_ERR_SiteAdmin_INVALID_SESSION':
            raise SessionError('The session expired or is invalid')


def check_request_status(data):
    if data['Status'] == 'Failed':

        if data['Error'] == 'SRV_ERR_SiteAdmin_INVALID_SESSION':
            raise SessionError('The session expired or is invalid', data['Error'])

        elif data['Error'] == 'SRV_ERR_SiteAdmin_INVALID_CRIDENTIALS':
            raise SessionError('The user name or password are wrong', data['Error'])

        elif data['Error'] == 'SRV_ERR_SiteAdmin_UNHANDLED_EXCEPTION':
            raise SessionError(data['ErrorAdmin'], data['Error'])

        """
        if 'NODES' not in data:
            raise RequestError('There are no NODES in the response')

        if 'ResultNode' not in data['NODES']:
            raise NoResultError('There are no ResultNodes in the response')
        """


class ResultSet(object):
    def __init__(self, data_string, start=0, end=0):
        self.data = json.loads(data_string, object_pairs_hook=OrderedDict)
        check_request_status(self.data)

        '''Indexing the nodes to find them faster using the method get_node.'''
        if 'NODES' in self.data and isinstance(self.data['NODES'], dict):
            self.data['NODES']['indexed_nodes'] = {}
            try:
                for index, node in enumerate(self.data['NODES']['Node']):
                    self.data['NODES']['indexed_nodes'][node['$URL']] = index
            except KeyError:
                pass

        '''Setting query result properties.'''
        self.start_index = start
        self.end_index = end
        self.total_count = int(self.data.get('TotalCount')) if self.data.get('TotalCount') is not None else 0
        self.has_more_results = self.data.get('HasMoreResults') == 'true'

    def result_nodes(self):
        if 'ResultNode' in self.data['NODES']:
            return self.data['NODES']['ResultNode']
        else:
            return []

    def get_node(self, url):
        try:
            return self.data['NODES']['Node'][self.data['NODES']['indexed_nodes'][url]]
        except TypeError:
            # Node with the requested url is not in data.
            return None


class Client(object):
    def __init__(self, host=None):
        """
        Initializer.
        :param str host: The host for the session.
        :rtype: None
        """
        self.uri = '/csi-requesthandler/RequestHandler?'
        self.host = host

        self.user = None
        self.password = None
        self.session_url = None
        self.security_token_url = None
        self.user_url = None

    @property
    def is_logged_in(self):
        """
        Check if there is an active session.
        :rtype: bool
        """
        return self.session_url is not None or self.security_token_url is not None

    @property
    def url(self):
        """
        Get the server endpoint that accepts requests.
        :rtype: str
        """
        return "{0}{1}".format(self.host, self.uri)

    def get_param_string(self, parameters, sign=True):
        """
        Generate a parameter string.
        :param parameters: The request parameters.
        :param sign: Should the session url be added to the request.
        :type parameters: dict of str
        :type sign: bool
        :rtype: str
        """
        items = []
        for key in parameters:
            value = parameters[key]
            if isinstance(value, list):
                for item_value in value:
                    items.append(urlencode({key: item_value}))
            else:
                items.append(urlencode({key: value}))

        if sign is True:
            if self.session_url is not None:
                items.append(urlencode({'SessionURL': self.session_url}))

            elif self.security_token_url is not None:
                items.append(urlencode({'SecurityTokenURL': self.security_token_url}))

        return '&'.join(items)

    def create_session_with_session_url(self, session_url):
        """
        Create a new session using a session url.
        :param session_url: The session url.
        :rtype: None
        """
        self.session_url = session_url
        self.validate_security_token()

    def create_session_with_security_token(self, security_token_url):
        """
        Create a new session using a security token.
        :param str security_token_url: The security token.
        :return:
        """
        self.security_token_url = security_token_url
        self.validate_security_token()

    def create_session_with_user(self, user, password):
        """
        Create a new session.
        :param str user: The user name.
        :param str password: The password.
        :rtype: None
        """
        self.user = user
        self.password = password
        self.session_url = None
        self.security_token_url = None
        self.user_url = None

        request_parameters = {
            'Module': 'DataSource',
            'Operation': 'SimpleLogin',
            'LoginID': self.user,
            'Password': self.password,
            'Fmt.Node.Info': 'Min',
            'OutputJSON': 2,
        }

        try:
            data = self.get_param_string(request_parameters, sign=False)
            result = urlopen(self.url, data.encode('UTF8')).read()
            data = load_json(result)
            check_request_status(data)

            self.session_url = data['NODES']['ResultNode'][0]['$URL']
            self.security_token_url = data['NODES']['ResultNode'][0]['SecurityToken']
            self.user_url = data['NODES']['ResultNode'][0]['UserInfo']

        except URLError:
            raise SessionError('Host is not reachable')

    @authenticate
    def destroy_session(self):
        """
        Destroy the active session.
        :rtype: None
        """

        if self.session_url is None or self.security_token_url is None:
            # If the session values do not exist, then the session cannot be destroyed.
            raise SessionError('The session expired or is invalid')

        request_parameters = {
            'Module': 'DataSource',
            'Operation': 'SimpleLogout',
            'Fmt.Node.Info': 'Min',
            'OutputJSON': 2,
        }
        data = self.do_request(request_parameters)

        # Unset the session values.
        self.session_url = None
        self.security_token_url = None
        self.user_url = None

    @authenticate
    def get_reflection(self, type):
        """
        Get the reflection data for a type
        :param type: string type of the class
        :return: dict
        """

        url = "{0}/csi-requesthandler/v1/reflection/classes/{1}?SessionURL={2}".format(self.host, type,
                                                                                       self.session_url)
        request = Request(url)

        request.add_header('Accept-Encoding', 'gzip')
        request.add_header('Accept', 'Accept	application/json, text/javascript, */*')

        try:
            response = urlopen(request)
        except HTTPError:
            raise NoResultError('A reflection class of this type does not exists.')

        if response.info().get('Content-Encoding') == 'gzip':
            '''http://stackoverflow.com/questions/3947120/does-python-urllib2-automatically-uncompress-gzip-data-fetched-from-webpage'''
            data = response.read()

            try:
                # python2
                buf = StringIO(data)
                f = gzip.GzipFile(fileobj=buf)
                data = f.read()
            except TypeError:
                # python3
                data = gzip.decompress(data).decode('UTF8')
        else:
            data = response.read()

        result = json.loads(data)

        return result

    @authenticate
    def do_request(self, parameters):
        """
        Make a server request.
        :param dict parameters:
        :rtype: str
        """
        request = Request(self.url,
                          data=self.get_param_string(parameters, sign=True).encode('UTF8'))

        request.add_header('Accept-Encoding', 'gzip')

        response = urlopen(request)

        if response.info().get('Content-Encoding') == 'gzip':
            '''http://stackoverflow.com/questions/3947120/does-python-urllib2-automatically-uncompress-gzip-data-fetched-from-webpage'''
            data = response.read()

            try:
                # python2
                buf = StringIO(data)
                f = gzip.GzipFile(fileobj=buf)
                data = f.read()
            except TypeError:
                # python3
                data = gzip.decompress(data).decode('UTF8')
        else:
            data = response.read()

        return data

    @authenticate
    def validate_session(self):
        """
        Validate the session by querying for the current user.
        :rtype: bool
        """
        if self.user_url is None:
            raise SessionError('User url is None')

        request_parameters = {
            'Fmt.Attr.Inc': 'No',
            'Fmt.Node.Info': 'Min',
        }
        self.query_by_url(self.user_url, parameters=request_parameters)

        # Will not get here if there is an exception.
        return True

    @authenticate
    def validate_security_token(self):
        """
        Verify that a security token is still valid. If it is, then populate session_url and user_url.
        from the response.
        :rtype: bool
        """
        request_parameters = {
            'Module': 'DataSource',
            'Operation': 'GetCurrentSession',
            'OutputJSON': 2,
        }

        try:
            result = urlopen(self.url, self.get_param_string(request_parameters, sign=True).encode('UTF8')).read()
            data = load_json(result)
            check_request_status(data)
            self.session_url = data['NODES']['ResultNode'][0]['$URL']
            self.user_url = data['NODES']['ResultNode'][0]['UserInfo']
            return True

        except URLError:
            raise SessionError('Host is not reachable')

    @authenticate
    def query_by_xml(self, xml, parameters=None):
        """
        Execute an xml query.
        :param str xml: The query in XML format.
        :param dict parameters: Additional parameters for the request.
        :rtype: ResultSet
        """
        query = '<?xml version="1.0" encoding="utf-8" ?><QUERY>{0}</QUERY>'.format(xml)

        request_parameters = {
            'Module': 'Search',
            'Operation': 'QueryByXML',
            'Fmt.Node.Info': 'Min',
            'Fmt.Attr.Inc': 'Yes',
            'Fmt.Attr.Info': 'ValueOnly',
            'Fmt.Attr.Format': 'Yes',
            'Fmt.Complete': 'Count',
            'Qry.XML': query,
            'Qry.Limit.Begin': 1,
            'Qry.Limit.End': 10,
            'OutputJSON': 2,
        }

        # Merge additionalParameters with params. These can be used to overwrite default param values.
        if parameters is not None:
            request_parameters.update(parameters)

        data = self.do_request(request_parameters)
        return ResultSet(data, start=request_parameters['Qry.Limit.Begin'], end=request_parameters['Qry.Limit.End'])

    @authenticate
    def query_by_url(self, url, parameters=None):
        """
        Get a node using the node's url.
        :param str url: The url of the node.
        :param dict parameters: Additional parameters for the request.
        :rtype: ResultSet
        """
        request_parameters = {
            'Module': 'Search',
            'Operation': 'QueryByURL',
            'Qry.URL': url,
            'Fmt.Node.Info': 'Max',
            'Fmt.Attr.Inc': 'Yes',
            'Fmt.Attr.Info': 'ValueOnly',
            'Fmt.Attr.Format': 'Yes',
            'Fmt.Complete': 'Count',
            'Qry.Limit.Begin': 1,
            'Qry.Limit.End': 10,
            'OutputJSON': 2,
        }

        # Merge additionalParameters with params. These can be used to overwrite default param values.
        if parameters is not None:
            request_parameters.update(parameters)

        data = self.do_request(request_parameters)
        return ResultSet(data, start=request_parameters['Qry.Limit.Begin'], end=request_parameters['Qry.Limit.End'])

    @authenticate
    def execute_xml_operations(self, xml, parameters=None):
        # TODO: add tests.
        """
        Execute an xml node operation string.
        :param xml: The node operation string in XML format.
        :param dict parameters: Additional parameters for the request.
        :rtype: ResultSet
        """
        query = '<?xml version="1.0" encoding="utf-8" ?>{0}'.format(xml)

        request_parameters = {
            'Module': 'NodeProcessor',
            'Operation': 'Execute',
            'OmitResults': 'false',
            'ReturnNodes': 'false',
            'Report': 'false',
            'UpdatedNodes': query,
            'OutputJSON': 2,
        }

        if parameters is not None:
            request_parameters.update(parameters)

        data = self.do_request(request_parameters)
        return ResultSet(data)

    @authenticate
    def get_image_src(self, image_url, size='SmallImage'):
        """
        Create a URL for an image that is signed with the current session.
        :param str image_url: The url of the image.
        :param str size: The size of the image.
        :rtype: str
        """
        if size not in ['SmallImage', 'Thumbnail', 'Viewable', ]:
            size = 'SmallImage'

        parameters = {
            'Module': 'Publisher',
            'Operation': 'GetFromNode',
            'OutputJSON': 1,
            'URL': image_url,
            'Attribute': size
        }

        return "{0}{1}".format(self.url, self.get_param_string(parameters, sign=True))

    @authenticate
    def get_file_src(self, file_url):
        """
        Create a URL for a file that is signed with the current session.
        :param str file_url: The url of the image.
        :rtype: str
        """

        parameters = {
            'Module': 'Publisher',
            'Operation': 'GetDirect',
            'URL': file_url,
        }

        return "{0}{1}".format(self.url, self.get_param_string(parameters, sign=True))

    @authenticate
    def generate_urls(self, type_=None, count=1):
        """
        Generate new urls on the server.
        ref: http://c8ipad.centricsoftware.com/csi-requesthandler/RequestHandler?Module=NodeProcessor&Operation=CreateNodeURL&NodeType=Material&Count=5

        :param str type_: The node type.
        :param int count: The number of urls to generate.
        :rtype: list of str
        """
        # TODO: Update unit tests.
        if type_ is None:
            return False

        parameters = {
            'Module': 'NodeProcessor',
            'Operation': 'CreateNodeURL',
            'NodeType': type_,
            'OutputJSON': 2,
            'Count': count
        }

        data = json.loads(self.do_request(parameters))
        if 'URL' in data['URLS']:
            return data['URLS']['URL']
        else:
            return []

    @authenticate
    def upload_file(self, document_url, file_name, file_path, parameters={}):
        """
        Uploads a file to a parent document contatiner.
        :param str document_url: The url of the document.
        :param str file_name: The name of the file.
        :param str file_path: The path to the file on the local disk.
        :param dict parameters: Additional parameters for the request.
        :return:
        """

        params = {
            'Module': 'Publisher',
            'Operation': 'Publish',
            'FN': file_name,
            'OutputJSON': 2,
        }
        params.update(parameters)
        params.update({'SessionURL': self.session_url})

        # Adding 'rb' for reading binary.
        try:
            file = open(file_path, 'rb')
            files = {'FL': file}  # 'file' => name of html input field
            data = requests.post(self.url, data=params, files=files)
        except UnicodeDecodeError:
            # Python 3.
            with open(file_path, 'rb') as f:
                files = {'FL': f}  # 'file' => name of html input field
                data = requests.post(self.url, data=params, files=files)
        finally:
            file.close()

        result = load_json(data.content)

        root = get_operations_root()
        change_node = ChangeNode(url="{0}?Path=CurrentRevision".format(document_url))
        change_node.add_change_attribute(ChangeAttribute(id='File', type='url', value=result['FileInfo']['URL']))
        change_node.add_change_attribute(
            ChangeAttribute(id='File.File', type='string', value=result['FileInfo']['FN']))
        root.add_change_node(change_node)
        result = self.execute_xml_operations(xml_to_string(root))
        return result

    @authenticate
    def create_document(self, parent_url, document_name):
        """
        Create a document node.
        :param parent_url: the url of the document and comment container. For example a style.
        :param document_name:
        :return:
        """
        root = get_operations_root()

        # Generate a url on the server for the document.
        url = self.generate_urls('Document', 1)

        # Add create the document node operation.
        create_node = CreateNode(type='Document')
        create_node.name = document_name
        create_node.url = url
        root.add_create_node(create_node)
        root.add_validate_node(url=url)

        # Add the operation to append the document url to the style.
        change_node = ChangeNode(url="{0}?Path=DocumentsAndComments".format(parent_url))
        change_attribute = ChangeAttribute(id='Documents', type='reflist', value_op='Append')
        change_attribute.add_ref(url)
        change_node.add_change_attribute(change_attribute)
        root.add_change_node(change_node)
        root.add_validate_node(url="{0}?Path=DocumentsAndComments".format(parent_url))

        # Execute the operations.
        request_parameters = {
            'OmitResults': 'false',
            'ReturnNodes': 'true',
        }
        result = self.execute_xml_operations(xml_to_string(root), request_parameters)

        # Find the document revision in the results.
        doc_rev_url = ''
        for node in result.data['NODES']['Node']:
            if node.get('$Type') == 'Document':
                return node

    @authenticate
    def revise_document(self, document_url):
        # TODO: Add tests.
        xml = """
        <NODES_OPERATIONS>
            <Call Module="SiteAdmin" Operation="ReviseAndAbandon">
                <Parameter Id="URL" Type="ref" Value="{0}?Path=CurrentRevision"/>
                <Parameter Id="ParentURL" Type="ref" Value="{0}"/>
                <Parameter Id="ParentAttr" Type="string" Value="Revisions"/>
                <Parameter Id="Toggle" Type="boolean" Value="true"/>
            </Call>
        </NODES_OPERATIONS>
        """.format(document_url)

        # Execute the operations.
        request_parameters = {
            'OmitResults': 'false',
            'ReturnNodes': 'true',
        }
        result = self.execute_xml_operations(xml, request_parameters)

        # Find the document revision in the results.
        doc_rev_url = ''
        for node in result.data['NODES']['Node']:
            if node.get('$Type') == 'DocumentRevision':
                return node

    @authenticate
    def upload_image(self, parent_url, file_name, file_path, parameters={}):
        # TODO: Add tests.
        """
        Uploads animage to a parent image container. For example a ColorSpecification.
        :param str parent_url: The url of the parent image container.
        :param str file_name: The name of the file.
        :param str file_path: The path to the file on the local disk.
        :param dict parameters: Additional parameters for the request.
        :return:
        """
        # TODO: Update unit tests.

        params = {
            'Module': 'SiteAdmin',
            'Operation': 'UploadImage',
            'Name': file_name,
            'URL': parent_url,
            'OutputJSON': 2,
        }
        params.update(parameters)
        params.update({'SessionURL': self.session_url})

        # Adding 'rb' for reading binary.
        try:
            file = open(file_path)
            files = {'Image': file}  # 'file' => name of html input field
            data = requests.post(self.url, data=params, files=files)
        except UnicodeDecodeError:
            # Python 3.
            with open(file_path, 'rb') as f:
                files = {'Image': f}  # 'file' => name of html input field
                data = requests.post(self.url, data=params, files=files)
        finally:
            file.close()

        return load_json(data.content)

    @authenticate
    def get_type_localizations(self, parameters={}):
        """
        Get type and attribute localized names for a given language code.
        :param dict parameters: Additional parameters for the request.
        :rtype: dict
        """

        request_parameters = {
            'Module': 'SiteAdmin',
            'Operation': 'GetTypesLocalization',
            'TypeFilter': "Style",
            'ResolveParameterization': "true",
            'LanguageCode': "en",
            'Fmt.Attr.Info': 'ValueOnly',
            'OutputJSON': 2,
        }

        request_parameters.update(parameters)

        try:
            result = urlopen(self.url, self.get_param_string(request_parameters, sign=True).encode('UTF8')).read()
            data = load_json(result)
            return data

        except ValueError:
            if len(result) == 0:
                # the result string is ""
                return {}
            else:
                # TODO: improve the error handling
                pass
