from __future__ import absolute_import
from .config import HOST, USER, PASSWORD
from centric8python.client import Client, exists
from centric8python.exceptions import SessionError, NoResultError
from unittest import TestCase
from collections import OrderedDict

try:
    # For Python 3.0 and later
    from urllib.parse import urlparse, parse_qsl
except ImportError:
    # Fall back to Python 2
    from urlparse import urlparse, parse_qsl


class TestRequest(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_login_and_logout(self):
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)
        self.assertTrue(req.is_logged_in)
        self.assertTrue(req.validate_session())

        req.destroy_session()
        self.assertTrue(req.is_logged_in == False)

        # a session needs to exist in order to destroy its
        self.assertRaises(SessionError, req.validate_session)
        self.assertRaises(SessionError, req.destroy_session)

    def test_reuse_session(self):
        # Create a session
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        host_session = {
            'csi_username': USER,
            'csi_password': PASSWORD,
            'csi_session_url': req.session_url,
            'csi_user_url': req.user_url,
        }

        # Reuse the session
        req1 = Client(host=HOST)
        req1.create_session_with_session_url(session_url=req.session_url)
        self.assertTrue(req1.validate_session())
        self.assertTrue(req.session_url == req1.session_url)

        req.destroy_session()
        self.assertRaises(SessionError, req.destroy_session)
        self.assertRaises(SessionError, req.validate_session)
        self.assertRaises(SessionError, req1.validate_session)
        self.assertRaises(SessionError, req1.destroy_session)

    def test_reuse_security_token(self):
        # Create a session
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)
        self.assertTrue(req.validate_security_token())

        host_session = {
            'csi_username': USER,
            'csi_password': PASSWORD,
            'csi_session_url': req.session_url,
            'csi_user_url': req.user_url,
        }

        # Reuse the security token
        req1 = Client(host=HOST)
        req1.create_session_with_security_token(security_token_url=req.security_token_url)
        self.assertTrue(req1.validate_session())
        self.assertTrue(req1.session_url is not None)
        self.assertTrue(req1.user_url is not None)
        self.assertTrue(req.security_token_url == req1.security_token_url)

        req.destroy_session()
        self.assertRaises(SessionError, req.destroy_session)
        self.assertRaises(SessionError, req.validate_session)

        req1.destroy_session()
        self.assertRaises(SessionError, req1.validate_session)
        self.assertRaises(SessionError, req1.destroy_session)
        self.assertRaises(SessionError, req1.validate_security_token)

    def test_invalid_host(self):
        req = Client(host="https://invalid_host")
        self.assertRaises(SessionError, req.create_session_with_user, user=USER, password=PASSWORD)
        self.assertFalse(req.is_logged_in)

    def test_invalid_user(self):
        req = Client(host=HOST)
        self.assertRaises(SessionError, req.create_session_with_user, user='x', password='x')
        self.assertFalse(req.is_logged_in)

    def test_invalid_session(self):
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)
        self.assertTrue(req.is_logged_in == True)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        '''

        parameters = {'Qry.Limit.Begin': 1, 'Qry.Limit.End': 1, }

        session_url = req.session_url
        req.session_url = 'centric://SESSION/86191652c0a80ae23967aafe700e8e00'

        def get_result():
            return req.query_by_xml(query, parameters=parameters)

        self.assertRaises(SessionError, get_result)

        req.session_url = session_url
        req.destroy_session()

    def test_get_param_string(self):
        req = Client(host=HOST)
        xml = ''
        query = '<?xml version="1.0" encoding="utf-8" ?><QUERY>' + xml + '</QUERY>'
        params = {
            'Module': 'Search',
            'Operation': 'QueryByXML',
            'Fmt.Attr.Inc': 'Yes',
            'Fmt.Node.Info': 'Max',
            'Fmt.Complete': 'Count',
            'Qry.XML': query
        }
        additional_data = None

        expected_result = 'Fmt.Attr.Inc=Yes&Qry.XML=%3C%3Fxml+version%3D%221.0%22+encoding%3D%22utf-8%22+%3F%3E%3CQUERY%3E%3C%2FQUERY%3E&Module=Search&Fmt.Node.Info=Max&Fmt.Complete=Count&Operation=QueryByXML'
        result = req.get_param_string(params, additional_data)
        self.assertTrue(len(result) > 0)

        a = {t[0]: t[1:] for t in parse_qsl(result)}
        b = {t[0]: t[1:] for t in parse_qsl(expected_result)}

        self.assertEqual(a, b)

    def test_get_invalid_styles(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        <Attribute Id="Node Name" Op="EQ" Value="9991212123" />
        '''

        parameters = {'Qry.Limit.Begin': 1, 'Qry.Limit.End': 10, }
        result = req.query_by_xml(query, parameters=parameters)
        self.assertTrue(result.total_count == 0)

        self.assertEqual(result.result_nodes(), [])
        self.assertEqual(result.get_node('does_not_exist'), None)

    def test_get_styles(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        '''

        parameters = {'Qry.Limit.Begin': 4,
                      'Qry.Limit.End': 6,
                      'Dep.Path': ['__Parent__', 'Attributes;Images;']
                      }

        # Make the request, and do not save the response.
        result = req.query_by_xml(query, parameters=parameters)

        self.assertTrue(len(result.result_nodes()) == 3)
        self.assertTrue(result.start_index == 4)
        self.assertTrue(result.end_index == 6)
        self.assertTrue(result.has_more_results)

        for node in result.result_nodes():
            self.assertTrue(result.get_node(node.get('__Parent__')) is not None)
            self.assertTrue(result.get_node(node.get('Attributes')) is not None)
            self.assertTrue(type(node.get('Images')) is OrderedDict)
            self.assertTrue(type(node.get('ProductSizes')) is list)
            self.assertTrue(type(node.get('DataSheets')) is list)

            # Is the image file accessible?
            self.assertTrue(exists(req.get_image_src(node.get('Images').get(''))))

        req.destroy_session()

    def test_get_type_localizations(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        parameters = {'TypeFilter': "Sales.*|Marketing.*|SM.*|MC.*|DocumentContainer.*"}
        result = req.get_type_localizations(parameters=parameters)

        self.assertTrue("MCMaterial" in result)
        self.assertTrue("MCMaterialColor" in result)
        self.assertTrue("DocumentContainer" in result)
        self.assertTrue("MCProduct" in result)
        self.assertTrue("MCProductColor" in result)
        self.assertTrue("MCProductToCustomer" in result)

        self.assertTrue("Attr" in result['MCProduct'])
        self.assertTrue("Code" in result['MCProduct']['Attr'])

        req.destroy_session()

    def test_get_type_localizations_invalid_options(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        parameters = {'TypeFilter': "InvalidClass"}
        result = req.get_type_localizations(parameters=parameters)
        self.assertTrue(len(result) == 0)

        req.destroy_session()

    def test_get_reflection(self):
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        model = req.get_reflection('Style')
        self.assertTrue(model.get('name') == 'Style')
        self.assertTrue(len(model.get('attributes')) > 0)

        model = req.get_reflection('Material')
        self.assertTrue(model.get('name') == 'Material')
        self.assertTrue(len(model.get('attributes')) > 0)

        req.destroy_session()

    def test_get_reflection_invalid_type(self):
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        self.assertRaises(NoResultError, req.get_reflection, 'XXStyle')

        req.destroy_session()

    def test_get_file_src(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        <Attribute Id="Images" Op="NN"/>
        '''

        parameters = {'Qry.Limit.Begin': 1,
                      'Qry.Limit.End': 1,
                      'Dep.Path': ['Images;']
                      }

        # Make the request, and do not save the response.
        result = req.query_by_xml(query, parameters=parameters)

        style = result.result_nodes()[0]
        image = result.get_node(style.get("Images")[""])

        # Check that the generated link for the file exists on the server.
        self.assertTrue(exists(req.get_file_src(image.get('SmallImage'))))

        req.destroy_session()
