from __future__ import absolute_import
import os
from unittest import TestCase
from centric8python.client import Client
from .config import HOST, USER, PASSWORD


class TestRequest(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_upload_document(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        '''

        parameters = {'Qry.Limit.Begin': 1,
                      'Qry.Limit.End': 1,
                      }

        # Make the request, and do not save the response.
        result = req.query_by_xml(query, parameters=parameters)
        style = result.result_nodes()[0]

        # Set the properties.
        document_name = "Example pdf Document"
        file_name = 'example.pdf'
        file_path = "{0}/{1}".format(os.path.dirname(os.path.realpath(__file__)), file_name)

        # Create the document container.
        document_node = req.create_document(style.get('$URL'), document_name)
        self.assertTrue('$URL' in document_node)
        self.assertTrue(document_node.get('$Name') == document_name)
        self.assertTrue(len(document_node.get('Revisions')) == 1)

        # Upload the file to the document container.
        result = req.upload_file(document_node.get('$URL'), file_name, file_path)

        # Check the data created on the server.
        parameters = {'Dep.Path': ['__Parent__', 'CurrentRevision']
                      }
        result2 = req.query_by_url(document_node.get('$URL'), parameters)

        document2 = result2.result_nodes()[0]
        document_container2 = result2.get_node(document2.get('__Parent__'))
        document_revision2 = result2.get_node(document2.get('CurrentRevision'))

        self.assertTrue(document_container2.get('__Parent__') == style.get('$URL'))
        self.assertTrue(document_revision2.get('File.File') == file_name)
        self.assertTrue(document_revision2.get('State') == 'Revision.State:DRAFT')

        # Revise the document.
        req.revise_document(document_node.get('$URL'))
        result = req.upload_file(document_node.get('$URL'), file_name, file_path)

        parameters = {'Dep.Path': ['__Parent__', 'Revisions']}
        result3 = req.query_by_url(document_node.get('$URL'), parameters)

        document3 = result3.result_nodes()[0]
        document_container3 = result3.get_node(document3.get('__Parent__'))
        document_revision3 = result3.get_node(document3.get('CurrentRevision'))
        document_revision2a = result3.get_node(document_revision2.get('$URL'))

        self.assertTrue(document_revision2a.get('State') == 'REVISED')

        self.assertTrue(document_container3.get('__Parent__') == style.get('$URL'))
        self.assertTrue(document_revision3.get('File.File') == file_name)
        self.assertTrue(document_revision3.get('State') == 'Revision.State:DRAFT')

        # Delete the document.
        xml = "<NODES_OPERATIONS><DeleteNode URL=\"{document_url}\"/></NODES_OPERATIONS>".format(document_url=document_node.get('$URL'))
        req.execute_xml_operations(xml)

        result4 = req.query_by_url(document_node.get('$URL'), parameters)
        self.assertTrue(len(result4.result_nodes()) == 0)

        req.destroy_session()

    def test_upload_image(self):
        # initialize request instance
        req = Client(host=HOST)
        req.create_session_with_user(user=USER, password=PASSWORD)

        # set up query
        query = '''
        <Node Parameter="Type" Op="EQ" Value="Style" />
        '''

        parameters = {'Qry.Limit.Begin': 1,
                      'Qry.Limit.End': 1,
                      }

        # Make the request, and do not save the response.
        result = req.query_by_xml(query, parameters=parameters)
        style = result.result_nodes()[0]

        # Set the properties.
        document_name = "Example jpeg image"
        file_name = 'example.jpg'

        file_path = "{0}/{1}".format(os.path.dirname(os.path.realpath(__file__)), file_name)
        # Upload the document.
        result = req.upload_image(style.get('$URL'), file_name, file_path)
        image_node = result['NODES']['ResultNode'][0]

        result = req.query_by_url(style.get('$URL'))
        style2 = result.result_nodes()[0]

        image_urls = style2.get('Images').values()
        self.assertTrue(image_node.get('$URL') in image_urls)

        # Delete the document.
        xml = "<NODES_OPERATIONS><DeleteNode URL=\"{image_url}\"/></NODES_OPERATIONS>".format(image_url=image_node.get('$URL'))
        req.execute_xml_operations(xml)

        result2 = req.query_by_url(image_node.get('$URL'), parameters)
        self.assertTrue(len(result2.result_nodes()) == 0)

        req.destroy_session()
